package es.optsicom.problem.gop.approx.improvement;

import java.util.Iterator;

import es.optsicom.lib.approx.improvement.AbstractImprovementMethod;
import es.optsicom.lib.util.ArraysUtil;
import es.optsicom.lib.util.CollectionsUtil;
import es.optsicom.lib.util.Id;
import es.optsicom.lib.util.Log;
import es.optsicom.lib.util.RandomList;
import es.optsicom.problem.gop.GOPInstance;
import es.optsicom.problem.gop.GOPSolution;
import es.optsicom.problem.gop.approx.improvement.OneVarLineSearch.ImprovementMode;
import es.optsicom.problem.gop.util.Util;

public class ScoredLineSearch extends AbstractImprovementMethod<GOPSolution, GOPInstance> {

	private IOneVarLineSearch lineSearch;

	public ScoredLineSearch() {
		this(new OneVarLineSearch(ImprovementMode.FIRST));
	}
	
	public ScoredLineSearch(IOneVarLineSearch oneVarLineSearch) {
		this.lineSearch = oneVarLineSearch;
	}
	
	@Override
	public boolean internalImproveSolution(GOPSolution solution, long millis) {
		GOPInstance instance = solution.getInstance();
		
		// annotate execution
		instance.incrementScoredLS();
		
		double lowerBound = instance.getFunction().getLowerBound();
		double upperBound = instance.getFunction().getUpperBound();
		
		double interval = (instance.getFunction().getUpperBound() - instance.getFunction().getLowerBound()) * instance.getStepSize();

		double[] weightValues = new double[instance.getDimension()];
		
		final int MAX_ITERATIONS = 10;
		int iterations = 0;
		boolean improved;
		do {
			
			// Determinar el grado de mejora aportado por cada variable
			GOPSolution workingCopy = (GOPSolution) solution.createCopy();
			for(int var = 0; var < instance.getDimension(); var++) {
				double oldValue = workingCopy.get(var);
				double leftValue = Util.makeMovement(oldValue, -interval, lowerBound, upperBound);
				workingCopy.set(var, leftValue);
				double left = workingCopy.getWeight();
				
				double rightValue = Util.makeMovement(oldValue, interval, lowerBound, upperBound);
				workingCopy.set(var, rightValue);
				double right = workingCopy.getWeight();
				
				if(left < right) {
					weightValues[var] = left;
				} else {
					weightValues[var] = right;
				}
				
				workingCopy.set(var, oldValue);
			}
			
			int[] order = ArraysUtil.sort(weightValues);

			double oldWeight = solution.getWeight();
			
			int numInspectedVariables = 0;
			while(numInspectedVariables <= solution.getInstance().getDimension() / 2) {
				lineSearch.performLineSearch(solution, order[numInspectedVariables], interval);
				numInspectedVariables++;
			}

			improved = oldWeight > solution.getWeight();
			
			if(solution.getWeight() > oldWeight) {
				throw new Error("Solution is worse after improvement");
			}
			
			iterations++;
			
		} while(improved && (iterations < MAX_ITERATIONS));
		
		return this.improveApplied;
	}
	
	@Override
	public boolean isDeterminist() {
		return false;
	}

	@Id
	public IOneVarLineSearch getLineSearch() {
		return lineSearch;
	}
}
