/* ******************************************************************************
 * 
 * This file is part of Optsicom
 * 
 * License:
 *   EPL: http://www.eclipse.org/legal/epl-v10.html
 *   LGPL 3.0: http://www.gnu.org/licenses/lgpl-3.0-standalone.html
 *   See the LICENSE file in the project's top-level directory for details.
 *
 * **************************************************************************** */
package es.optsicom.problem.gop.approx.pr;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import es.optsicom.lib.DistanceCalc;
import es.optsicom.lib.Instance;
import es.optsicom.lib.Solution;
import es.optsicom.lib.approx.AbstractApproxMethod;
import es.optsicom.lib.approx.algorithm.pr.EliteSet;
import es.optsicom.lib.approx.constructive.Constructive;
import es.optsicom.lib.approx.improvement.ImprovementMethod;
import es.optsicom.lib.util.ArraysUtil;
import es.optsicom.lib.util.CollectionsUtil;
import es.optsicom.lib.util.Id;
import es.optsicom.lib.util.Log;
import es.optsicom.lib.util.MathUtil;
import es.optsicom.lib.util.RandomManager;
import es.optsicom.lib.util.RandomizedSelector;
import es.optsicom.problem.gop.GOPInstance;
import es.optsicom.problem.gop.GOPSolution;

public abstract class AbstractPR extends AbstractApproxMethod<GOPSolution, GOPInstance> {

	protected int eliteSetSize = 10;
	protected Constructive<GOPSolution, GOPInstance> constructive;
	protected GOPPathRelinking pathRelinking;
	protected ImprovementMethod<GOPSolution, GOPInstance> improvement;
	protected DistanceCalc<GOPSolution, GOPInstance> distCalc;
	protected double thresold;

	protected EliteSet<GOPSolution, GOPInstance> es;
	private boolean considerDiversity;
	
	public AbstractPR() {
		
	}
	
	public AbstractPR(Constructive<GOPSolution, GOPInstance> constructive, GOPPathRelinking pathRelinking,
	        ImprovementMethod<GOPSolution, GOPInstance> improvement, DistanceCalc<GOPSolution, GOPInstance> distCalc, double thresold) {
		this.constructive = constructive;
		this.pathRelinking = pathRelinking;
		this.improvement = improvement;
		this.distCalc = distCalc;
		this.thresold = thresold;
		this.considerDiversity = true;
	}

	@Id
	public double getThresold() {
		return thresold;
	}
	
	public void setThresold(double thresold) {
		this.thresold = thresold;
	}

	@Id
	public Constructive<GOPSolution, GOPInstance> getConstructive() {
		return constructive;
	}

	@Id
	public GOPPathRelinking getPathRelinking() {
		return pathRelinking;
	}

	@Id
	public ImprovementMethod<GOPSolution, GOPInstance> getImprovement() {
		return improvement;
	}

	@Id
	public DistanceCalc<GOPSolution, GOPInstance> getDistCalc() {
		return distCalc;
	}

	protected void populateElisteSet(int numConstructions) {

		for (int i = 0; i < numConstructions; i++) {
			GOPSolution solution = constructive.createSolution();
			setIfBestSolution(solution);
			
			if(i == 0) {
				improvement.improveSolution(solution);
			}
			
//			improvement.improveSolution(solution);
//			setIfBestSolution(solution);

			es.addSolution(solution);
			
		}
	}

	protected void doDynEliteSetIteration() {

		GOPSolution solution = constructive.createSolution();
		setIfBestSolution(solution);
		
		// ¿Hay que mejorar aquí?
//		improvement.improveSolution(solution);
		setIfBestSolution(solution);

		Solution[] esSols = es.getSolutions();
		List<GOPSolution> remainingSolutions = new ArrayList<GOPSolution>();
		for(Solution s : esSols) {
			remainingSolutions.add((GOPSolution)s);
		}
		
		// Obtener dos soluciones proporcionales a la calidad
		double[] weights = new double[remainingSolutions.size()];
		for(int index = 0; index < weights.length; index++) {
			weights[index] = remainingSolutions.get(index).getWeight();
		}
		
		int index1 = MathUtil.selectIndexInversely(weights);
		if(index1 == weights.length) {
			index1--;
		}
		GOPSolution sol1 = remainingSolutions.remove(index1);
		
		weights = new double[remainingSolutions.size()];
		for(int index = 0; index < weights.length; index++) {
			weights[index] = remainingSolutions.get(index).getWeight();
		}
		
		int index2 = MathUtil.selectIndexInversely(weights);
		if(index2 == weights.length) {
			index2--;
		}
		GOPSolution sol2 = remainingSolutions.get(index2);
		
		List<GOPSolution> guidingSolutions = new ArrayList<GOPSolution>();
		guidingSolutions.add(sol1);
		guidingSolutions.add(sol2);
		
		GOPSolution prSol = this.pathRelinking.createSolution(solution, guidingSolutions);
		
		if (prSol != null) {
			improvement.improveSolution(prSol);
			setIfBestSolution(prSol);
			int index = es.addSolution(prSol);
			
//			if(index != -1) {
//				Log.debugln("Entra en el ES");
//			} else {
//				Log.debugln("No entra en el ES");
//			}
		}

	}
	
	//Es una versión simplificada del método doDynEliteSetIteration. Es simplificada porque 
	//la selección del la solución del ES se realiza de forma totalmente aleatoria, en vez vez
	//de realizarse de forma proporcional a la calidad. Además, también es una versión simplificada
	//porque no tiene control de diversidad al insertar una solución al ES. Simplemente añade una
	//solución si es mejor que la peor del ES. 
	protected void doSimpleDynEliteSetIteration() {

		//System.out.println("SimpleDynEliteSetIteration");
		
		GOPSolution solution = constructive.createSolution();
		setIfBestSolution(solution);
		improvement.improveSolution(solution);
		setIfBestSolution(solution);

		Solution[] esSols = es.getSolutions();
		List<GOPSolution> remainingSolutions = new ArrayList<GOPSolution>();
		for(Solution s : esSols) {
			remainingSolutions.add((GOPSolution)s);
		}
		List<GOPSolution> guidingSolutions = new ArrayList<GOPSolution>();
		do {
			int numSol = RandomManager.getRandom().nextInt(remainingSolutions.size());
			GOPSolution esSol = remainingSolutions.remove(numSol);
			if(!esSol.equals(solution)) {
				guidingSolutions.add(esSol);
			}
		} while (guidingSolutions.size() < 2 && !remainingSolutions.isEmpty());

		if(guidingSolutions.size() < 2) {
			throw new Error("Couldn't find different solutions");
		}
		
		GOPSolution prSol = this.pathRelinking.createSolution(solution, guidingSolutions);
		//System.out.println("PathRelinking: "+prSol.getWeight());
		if (prSol != null) {
			
			improvement.improveSolution(prSol);
			//System.out.println("ImprovedSolution: "+prSol.getWeight());
			setIfBestSolution(prSol);
			es.addSolutionWithoutDiversityCheck(prSol);
		}

	}

	@Id
	public int getEliteSetSize() {
		return eliteSetSize;
	}

	public void setEliteSetSize(int eliteSetSize) {
		this.eliteSetSize = eliteSetSize;
	}
	
	@Override
	public void setInstance(GOPInstance instance) {
		super.setInstance(instance);
		instance.resetEvaluations();
	}
	
	@Id
	public boolean isConsiderDiversity() {
		return considerDiversity;
	}
	
	public void setConsiderDiversity(boolean considerDiversity) {
		this.considerDiversity = considerDiversity;
	}

	public void setImprovement(ImprovementMethod<GOPSolution, GOPInstance> improvement) {
		this.improvement = improvement;
	}
	
	public void setPathRelinking(GOPPathRelinking pathRelinking) {
		this.pathRelinking = pathRelinking;
	}
	
	public void setConstructive(Constructive<GOPSolution, GOPInstance> constructive) {
		this.constructive = constructive;
	}
}