/* ******************************************************************************
 * 
 * This file is part of Optsicom
 * 
 * License:
 *   EPL: http://www.eclipse.org/legal/epl-v10.html
 *   LGPL 3.0: http://www.gnu.org/licenses/lgpl-3.0-standalone.html
 *   See the LICENSE file in the project's top-level directory for details.
 *
 * **************************************************************************** */
package es.optsicom.problem.gop.approx.pr;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import es.optsicom.lib.Solution;
import es.optsicom.lib.approx.algorithm.pr.EliteSet;
import es.optsicom.lib.experiment.CurrentExperiment;
import es.optsicom.lib.util.Id;
import es.optsicom.lib.util.Log;
import es.optsicom.problem.gop.GOPInstance;
import es.optsicom.problem.gop.GOPProblem;
import es.optsicom.problem.gop.GOPSolution;
import es.optsicom.problem.gop.MaxNumberEvaluationsExceededException;

public class EvolutionaryPathRelinkingMethod extends AbstractPR {

	private int globalIter = 20;
	private int localIter = 10;

	public EvolutionaryPathRelinkingMethod() {
		super();
	}
	
	@Id
	public int getGlobalIter() {
		return globalIter;
	}

	public EvolutionaryPathRelinkingMethod setGlobalIter(int iter) {
		globalIter = iter;
		return this;
	}

	@Id
	public int getLocalIter() {
		return localIter;
	}

	public EvolutionaryPathRelinkingMethod setLocalIter(int iter) {
		localIter = iter;
		return this;
	}

	@Override
	protected void internalCalculateSolution(long finishTime) {
		try {
			internalCalculateSolution2();
		} catch (MaxNumberEvaluationsExceededException e) {
			// Stop.
			Log.debugln("Stopping...");

			// Save LS calls
			if(GOPInstance.SAVE_EVALUATION_EVENTS) {
				CurrentExperiment.addEvent(GOPProblem.SCORED_LS_CALLS, this.getInstance().getScoredLSCalls());
			}
		}
	}
	
	protected void internalCalculateSolution2() {

		//		1. Set GlobalIter equal to the number of global iterations.
		//		2. Apply the GRASP method (construction plus improvement) for b=|ES| iterations to populate ES. ES={ x1, x2, …, xb }.
		//
		//		For(iter= 1 to GIter)
		//			For(i=1 to LocalIter)
		//				3. Apply the construction phase of GRASP  x.
		//				4. Apply the local search phase of GRASP to x  x'.
		//				5. Randomly select xj in ES (Map values into probabilities).
		//				6. Apply PR(x', xj) and PR(xj,x'), let y be the best solution found.
		//				7. Apply the local search phase of GRASP to y  y'.
		//				If ( zMM(y') > zMM(x1)  or  (zMM(y') > zMM(xb) and d(y',ES)  dth))
		//					8. Let xk be the closest solution to y' in ES with zMM(y')>zMM(xk).
		//					9. Add y' to ES and remove xk.
		//					10. Update the order in ES (from the best x1 to the worst xb).
		//
		//			11. NewSol=1
		//			While(NewSol) 
		//				12. NewSol=0
		//		13. Apply PR(x,x’) and PR(x’,x) for every pair (x,x’) in ES not combined before.  Let y be the best solution found.
		//		14. Apply the local search phase of GRASP to y  y'.
		//				If ( zMM(y') > zMM(x1)  or  (zMM(y') > zMM(xb) and d(y',ES)  dth))
		//					15. Let xk be the closest solution to y' in ES with zMM(y')>zMM(xk).
		//					16. Add y' to ES and remove xk.
		//					17. Update the order in ES (from the best x1 to the worst xb).
		//					18. NewSol=1
		//
		//		19. Return x1.

		this.constructive.setInstance(instance);

		this.es = new EliteSet<GOPSolution, GOPInstance>(eliteSetSize, distCalc, thresold * this.instance.getBoundsWidth());

		this.populateElisteSet(eliteSetSize);

		while(true) {

			for (int j = 0; j < localIter; j++) {
				if(isConsiderDiversity()) {
					doDynEliteSetIteration();
				} else {
					doSimpleDynEliteSetIteration();
				}
			}

			Log.debugln("-------------------------------------Global");
			eliteSetEvolution();
			
			globalFinished();
		}
		
	}

	protected void globalFinished() {
		// Do nothing
	}
	
	protected void eliteSetEvolution() {

		//System.out.println("Evolution ES");

		int newEsSolutionsCounter;

		final boolean NEW = true;
		final boolean OLD = false;
		
		int iterations = 0;

		boolean[] esSolutions = new boolean[eliteSetSize];
		Arrays.fill(esSolutions, NEW);

		do {

			newEsSolutionsCounter = 0;

			List<GOPSolution> solutions = new ArrayList<GOPSolution>();
			
			Solution[] esSols = es.getSolutions();
			List<GOPSolution> esSolsList = new ArrayList<GOPSolution>();
			for(Solution s : esSols) {
				esSolsList.add((GOPSolution)s);
			}

			
//			for (int i = 0; i < eliteSetSize; i++) {
//				for (int j = 0; j < eliteSetSize/2; j++) {
//					if(i == j || i == j+1) {
//						continue;
//					}
			for (int i = 0; i < eliteSetSize-2; i++) {
				for (int j = i + 1; j < eliteSetSize-1; j++) {
					int k = j+1;
					if (esSolutions[i] == NEW || esSolutions[j] == NEW || esSolutions[k] == NEW) {

						GOPSolution solI = es.getSolution(i);
						GOPSolution solJ = es.getSolution(j);
						GOPSolution solK = es.getSolution(k);
						
						if (!solI.equals(solJ) && !solJ.equals(solK) && !solK.equals(solI)) {

							GOPSolution prSol = this.pathRelinking.createSolution(solI, Arrays.asList(solJ, solK));

							if (prSol != null) {
								improvement.improveSolution(prSol);
								setIfBestSolution(prSol);

//								if(prSol.isBetterThan(solI) && prSol.isBetterThan(solJ) && prSol.isBetterThan(solK)) {
//									Log.debugln("New sol improved");
//								}

								solutions.add(prSol);
							}
						}
					}

				}
			}

			for (int i = 0; i < eliteSetSize; i++) {
				esSolutions[i] = OLD;
			}

			for (GOPSolution solution : solutions) {
				if(isConsiderDiversity()) {
					int index = es.addSolution(solution);
					if (index != -1) {
						esSolutions[index] = NEW;
						newEsSolutionsCounter++;
					}
				} else {
					boolean inserted = es.addSolutionWithoutDiversityCheck(solution);
					if(inserted) {
						esSolutions[eliteSetSize-1] = NEW;
						newEsSolutionsCounter++;
					}
				}
			}

			iterations++;
//			System.out.println(iterations);
			
		} while ((newEsSolutionsCounter > 0) && (iterations < globalIter));

	}
}
